﻿using Microsoft.Extensions.Configuration;
using Serilog;
using Spectre.Console;

namespace PhotoRenamer;

internal class Renamer
{
    private readonly string _sourcePath;
    private readonly string _targetPath;
    private int _currentCount;

    public Renamer(IConfiguration configuration)
    {
        _sourcePath = Path.GetFullPath(
            configuration["Source"]
                ?? throw new NotSupportedException("Source directory needs to be set.")
        );
        _targetPath = Path.GetFullPath(
            configuration["Target"]
                ?? throw new NotSupportedException("Target directory needs to be set.")
        );
        Log.Information($"Source path: {_sourcePath}");
        Log.Information($"Target path: {_targetPath}");
    }

    public async Task<int> RunAsync()
    {
        var files = FolderExtension.GetAllFiles(_sourcePath);
        _currentCount = 0;
        var po = new ParallelOptions { MaxDegreeOfParallelism = 3 };
        await AnsiConsole
            .Progress()
            .StartAsync(async ctx =>
            {
                await Parallel.ForEachAsync(files, po, (s, token) => Body(s, ctx, token));
            });
        return 0;
    }

    private async ValueTask Body(
        string file,
        ProgressContext progressContext,
        CancellationToken token
    )
    {
        if (token.IsCancellationRequested)
            return;
        var task = progressContext.AddTask($"[green]{file}[/]");
        try
        {
            var dateTime = MediaFileExtension.GetDateTimeFromSourceFile(file);
            var folder = FolderExtension.CreateFolderStructureByDate(
                _targetPath,
                DateOnly.FromDateTime(dateTime)
            );
            var progress = new Progress<double>(v => task.Value = v * 100);

            await CopyFileAsync(folder, file, progress);
        }
        catch (Exception ex)
        {
            Log.Error(ex, "Error reading file information");
        }
        finally
        {
            Interlocked.Increment(ref _currentCount);
            task.StopTask();
        }
    }

    private static void ResetTimes(FileSystemInfo destination, FileSystemInfo source)
    {
        destination.LastWriteTime = source.LastWriteTime;
        destination.LastWriteTimeUtc = source.LastWriteTimeUtc;
        destination.CreationTime = source.CreationTime;
        destination.CreationTimeUtc = source.CreationTimeUtc;
    }

    private static async Task CopyFileAsync(
        string folder,
        string file,
        IProgress<double>? progress = null
    )
    {
        var destination = new FileInfo(Path.Combine(folder, Path.GetFileName(file)));
        var source = new FileInfo(file);
        if (destination.Exists && destination.Length == source.Length)
        {
            progress?.Report(1);
            return;
        }

        await using var sourceStream = File.Open(
            source.FullName,
            FileMode.Open,
            FileAccess.Read,
            FileShare.Read
        );
        await using var targetStream = File.Open(
            destination.FullName,
            FileMode.Create,
            FileAccess.ReadWrite,
            FileShare.ReadWrite
        );

        await sourceStream.CopyToAsync(targetStream, 81920, progress: progress);

        ResetTimes(destination, source);
    }
}
